"""
Dummy.

"""
from typing import Any, Union
from collections.abc import Callable
from navconfig.logging import logging
from notify.utils import Msg
from notify.models import Actor
from notify.providers.abstract import ProviderBase, ProviderType


def dummy_sent(obj: ProviderBase, recipient: Actor, message: Union[str, Any], result: Any, **kwargs): # pylint: disable=W0613
    logging.debug(f'Message Sent! {recipient!s}')
    #


class Dummy(ProviderBase):
    """
    dummy.

    Dummy Provider to send messages to stdout
    """
    provider = 'dummy'
    provider_type = ProviderType.NOTIFY
    blocking: bool = True
    sent: Callable = dummy_sent

    async def connect(self, *args, **kwargs):
        print('Connecting to Dummy ...')

    async def close(self):
        print('Closing to Dummy...')

    async def _send_(self, to: Actor, message: Union[str, Any], subject: str = None, **kwargs):
        """
        _send.

        Logic associated with the construction of notifications
        """
        msg = await self._render_(to=to, message=message, **kwargs)
        try:
            Msg(str(msg))
        except TypeError:
            print(msg)
        return msg



from navconfig import config


SLACK_APP_ID = config.get('SLACK_APP_ID')
SLACK_CLIENT_ID = config.get('SLACK_CLIENT_ID')
SLACK_CLIENT_SECRET = config.get('SLACK_CLIENT_SECRET')
SLACK_SIGNING_SECRET = config.get('SLACK_SIGNING_SECRET')
# Bot information:
SLACK_BOT_TOKEN = config.get('SLACK_BOT_TOKEN')
SLACK_DEFAULT_CHANNEL = config.get('SLACK_DEFAULT_CHANNEL')
SLACK_TEAM_ID = config.get('SLACK_TEAM_ID')



# -*- coding: utf-8 -*-
from notify.providers.mail import ProviderEmail
from .settings import (
    SENDGRID_USER,
    SENDGRID_KEY
)

class Sendgrid(ProviderEmail):
    """
    Sendgrid.

    Sendgrid SMTP Client.
    TODO: migrate to API.
    """
    provider = 'sendgrid'
    blocking: bool = False


    def __init__(self, username: str = None, password: str = None, **kwargs):
        """

        """
        self._attachments: list = []
        self.force_tls: bool = True

        super(Sendgrid, self).__init__(**kwargs)

        # server information
        self._host = 'smtp.sendgrid.net'
        self._port = 587
        # connection related settings
        self.username = username
        if not self.username:
            self.username = SENDGRID_USER

        self.password = password
        if not self.password:
            self.password = SENDGRID_KEY

        try:
            # sent from another account
            if 'account' in kwargs:
                self.actor = kwargs['account']
            else:
                self.actor = self.username
        except KeyError:
            self.actor = self.username



"""
Twitter API.

Using tweepy to send (publish) Tweets.
"""
from typing import (
    Optional,
    Union,
    Any
)
import tweepy
from notify.providers import ProviderIMBase, ProviderType
from notify.models import Actor
from notify.exceptions import ProviderError
from .settings import TWITTER_ACCESS_TOKEN, TWITTER_TOKEN_SECRET, TWITTER_CONSUMER_KEY, TWITTER_CONSUMER_SECRET


class Twitter(ProviderIMBase):
    provider = 'twitter'
    provider_type = ProviderType.IM
    _consumer_key: str = None
    _consumer_secret: str = None
    _token: str = None
    _secret: str = None
    client = None

    sid = None

    def __init__(
            self,
            consumer_key=None,
            consumer_secret=None,
            token=None,
            secret=None,
            *args,
            **kwargs
        ):
        """
        :param token: twilio auth token given by Twilio
        :param sid: twilio auth id given by Twilio

        """
        super(Twitter, self).__init__(*args, **kwargs)
        self._token = TWITTER_ACCESS_TOKEN if token is None else token
        self._secret = TWITTER_TOKEN_SECRET if secret is None else kwargs['sid']
        self._consumer_key = TWITTER_CONSUMER_KEY if consumer_key is None else consumer_key
        self._consumer_secret = TWITTER_CONSUMER_SECRET if consumer_secret is None else consumer_secret

    def close(self):
        self.client = None

    def connect(self):
        """
        Verifies that a token and sid were given

        """
        if self._token is None or self._secret is None:
            raise RuntimeError(
                f'to send Tweets via {self._token} you need to configure TWITTER_ACCESS_TOKEN & TWITTER_TOKEN_SECRET in \n'
                'environment variables or send properties to theinstance.'
            )
        try:
            auth = tweepy.OAuthHandler(
                self._consumer_key,
                self._consumer_secret
            )
            if auth:
                auth.set_access_token(self._token, self._secret)
                self.client = tweepy.API(auth)
        except Exception as err:
            raise ProviderError(
                f'Twitter API error: {err}'
            ) from err
        # determine actor:
        self.actor = self.client.me()


    async def _send_(self, to: Optional[Actor] = None, message: Union[str, Any] = None, **kwargs) -> Any:
        """
        _send_.

        Publish a Twitter using Tweepy
        :param to: Optional Reply-To Message.
        :param message: message to Publish
        """
        try:
            msg = self._render(to, message, **kwargs)
            result = self.client.update_status(status=msg)
            # TODO: processing the output, adding callbacks
            return result
        except Exception as ex:
            print(ex)
            raise ProviderError(
                f'Error Publishing Tweet, Current Error: {ex}'
            ) from ex



"""Abstract.

Base Factory classes for all kind of Providers.
"""
import asyncio
import time
from abc import ABC, abstractmethod
from enum import Enum
from functools import partial
from concurrent.futures import ThreadPoolExecutor
from typing import (
    Any,
    Union
)
from collections.abc import Awaitable, Callable
from navconfig.logging import logging
from notify.utils import SafeDict, cPrint
from notify.exceptions import ProviderError, MessageError
from notify.models import Actor
from notify.settings import NAVCONFIG, DEBUG


class ProviderType(Enum):
    NOTIFY = 'notify'  # generic notification
    SMS = 'sms'  # SMS messages
    EMAIL = 'email'  # email (smtp) notifications
    PUSH = 'push'  # push notifications
    IM = 'im'  # instant messaging


class ProviderBase(ABC):
    """ProviderBase.

    Base class for All providers
    """
    provider: str = None
    provider_type: ProviderType = ProviderType.NOTIFY
    blocking: bool = True
    sent: Callable = None

    def __init__(self, **kwargs):
        self.__name__ = str(self.__class__.__name__)
        self._params = kwargs
        self._logger = logging.getLogger('Notify')
        # environment config
        self._config = NAVCONFIG
        if 'loop' in kwargs:
            self._loop = kwargs['loop']
            del kwargs['loop']
        else:
            self._loop = asyncio.get_event_loop()
        asyncio.set_event_loop(self._loop)
        # configure debug:
        if 'debug' in kwargs:
            self._debug = kwargs['debug']
            del kwargs['debug']
        else:
            self._debug = DEBUG
        # add the Jinja Template Parser
        try:
            from notify import TemplateEnv # pylint: disable=C0415
            self._tpl = TemplateEnv
            self._template = None
        except Exception as err:
            raise RuntimeError(
                f"Notify: Can't load the Jinja2 Template Parser: {err}"
            ) from err
        # set the values of attributes:
        for arg, val in self._params.items():
            try:
                object.__setattr__(self, arg, val)
            except AttributeError:
                pass


### Async Context magic Methods
    async def __aenter__(self) -> "ProviderBase":
        if asyncio.iscoroutinefunction(self.connect):
            await self.connect()
        else:
            self.connect()
        return self

    def __enter__(self) -> "ProviderBase":
        self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        # clean up anything you need to clean up
        try:
            if asyncio.iscoroutinefunction(self.close):
                await self.close()
            else:
                self.close()
        except Exception as err: # pylint: disable=W0703
            logging.error(err)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        try:
            self.close()
        except Exception as err: # pylint: disable=W0703
            logging.error(err)

    @abstractmethod
    async def connect(self, *args, **kwargs):
        pass

    @abstractmethod
    async def close(self):
        pass

    @abstractmethod
    async def _send_(self, to: Actor, message: Union[str, Any], subject: str = None, **kwargs) -> Any:
        """_send_.
          Method called for every recipient in Recipient list.
        Args:
            to (Actor): recipient of message.
            message (Union[str, Any]): message data.

        Raises:
            RuntimeError: Error when _send_ can be executed.

        Returns:
            Any: Result of sending process.
        """

    @classmethod
    def name(cls):
        return cls.__name__

    @classmethod
    def type(cls):
        return cls.provider_type

    def get_loop(self):
        return self._loop

    def set_loop(self, loop: asyncio.AbstractEventLoop = None):
        if not loop:
            self._loop = asyncio.new_event_loop()
        else:
            self._loop = loop
        asyncio.set_event_loop(self._loop)

    async def _prepare_(
            self,
            recipient: Actor = None,
            message: Union[str, Any] = None,
            template: str = None,
            **kwargs
        ): # pylint: disable=W0613
        """
        _prepare.

        Prepare a Message for Sending.
        """
        #1 replacement of strings
        if self._params:
            msg = message.format_map(
                SafeDict(**self._params)
            )
        else:
            msg = message
        if template:
            # using a template parser:
            self._template = self._tpl.get_template(template)
        else:
            self._template = None
        return msg

    async def _render_(
            self,
            to: Actor = None,
            message: str = None,
            subject: str = None,
            **kwargs
        ): # pylint: disable=W0613
        """
        _render_.

        Returns the parseable version of Message template.
        """
        # cPrint(f'RECEIVED {to}, message {message}')
        msg = message
        if self._template:
            self._templateargs = {
                "recipient": to,
                "username": to,
                "message": message,
                **kwargs
            }
            msg = self._template.render(**self._templateargs)
        return msg

    def create_task(self, to, message, loop: asyncio.AbstractEventLoop, subject: str = None, **kwargs):
        asyncio.set_event_loop(loop)
        task = loop.create_task(self._send_(to, message, subject=subject, **kwargs))
        fn = partial(self.__sent__, to, message, **kwargs)
        task.add_done_callback(fn)
        return task

    async def send(
            self,
            recipient: list[Actor] = None,
            message: Union[str, Any] = None,
            subject: str = None,
            **kwargs
        ):
        """
        send.

        public method to send messages and notifications
        """
        # template (or message) for preparation
        msg = await self._prepare_(
            recipient=recipient,
            message=message,
            **kwargs
        )
        rcpt = []
        results = None
        if isinstance(recipient, list):
            rcpt = recipient
        else:
            rcpt.append(recipient)
        # TODO: non-blocking code
        if self.blocking is True:
            loop = asyncio.new_event_loop()
            tasks = []
            for to in rcpt:
                task = self.create_task(to, msg, loop, subject, **kwargs)
                tasks.append(task)
            # creating the executor
            fn = partial(
                self.execute_notify, loop, tasks
            )
            with ThreadPoolExecutor(max_workers=len(tasks)) as pool:
                results = loop.run_in_executor(pool, fn)
        else:
            asyncio.set_event_loop(self._loop)
            queue = asyncio.Queue(maxsize=len(rcpt)+1)
            started_at = time.monotonic()
            tasks = []
            consumers = []
            i = 0
            for to in rcpt:
                # create the consumers:
                consumer = asyncio.create_task(
                    self.process_notify(queue)
                )
                consumers.append(consumer)
                # create the task:
                task = self.create_task(to, message, self._loop, subject, **kwargs)
                tasks.append(task)
                i += 1
            # send tasks to queue processor (producer)
            await asyncio.gather(*[self.notify_producer(queue, tasks)])
            # wait until the consumer has processed all items
            await queue.join()
            total_slept = time.monotonic() - started_at
            # Cancel our worker tasks.
            for task in consumers:
                task.cancel()
            print(f'System ended: {total_slept}')

        return results

    # PUB/SUB Logic based on Asyncio Queues
    async def notify_producer(self, queue, tasks: list):
        """
        Process Notify.
        Fill the asyncio Queue with tasks
        """
        for task in tasks:
            await queue.put(task)

    async def process_notify(self, queue):
        """
        Consumer logic of Asyncio Queue
        """
        while True:
            # Get a "work item" out of the queue.
            task = await queue.get()
            # process the task
            try:
                await task
            except (ProviderError, MessageError) as ex:
                logging.error(ex)
                raise
            finally:
                # Notify the queue that the item has been processed
                queue.task_done()

    def execute_notify(
            self,
            loop: asyncio.AbstractEventLoop,
            tasks: list[Awaitable]
            ):
        """
        execute_notify.
        Executing notification in a event loop.
        """
        asyncio.set_event_loop(loop)
        try:
            group = asyncio.gather(*tasks, return_exceptions=True)
            try:
                results = loop.run_until_complete(group)
                return results
            except RuntimeError as err:
                raise RuntimeError(
                    f"Notify: Error executing Notify: {err}"
                ) from err
            except Exception as err:
                logging.exception(err, stack_info=True)
                raise ProviderError(
                    f"Notify: Error executing Notify: {err}"
                ) from err
        except Exception as err:
            raise ProviderError(
                f"Notify: Exception on Notify: {err}"
            ) from err

    def __sent__(
                self,
                recipient: Actor,
                message: str,
                _task: Awaitable,
                **kwargs
            ):
        """
        processing the callback for every notification that we sent.
        """
        if callable(self.sent):
            try:
                result = _task.result()
                # logging:
                self._logger.debug(
                    f'Notification sent to:> {recipient}'
                )
                evt = asyncio.new_event_loop()
                if 'task' in kwargs:
                    del kwargs['task']
                fn = partial(self.callback_sent, recipient, message, result, evt, **kwargs)
                with ThreadPoolExecutor(max_workers=1) as executor:
                    evt.run_in_executor(executor, fn)
            except (AttributeError, RuntimeError) as ex:
                self._logger.exception(
                    f'Notify: *Sent* Function fail with error {ex}'
                )

    def callback_sent(
            self,
            recipient: Actor,
            message: Any,
            result: Any,
            evt: asyncio.AbstractEventLoop,
            **kwargs
        ) -> None:
        """callback_sent.

        Function for running Callback in a Thread.
        """
        asyncio.set_event_loop(evt)
        fn = partial(self.sent, recipient, message, result, **kwargs)
        try:
            if asyncio.iscoroutinefunction(self.sent):
                evt.run_until_complete(fn)
            else:
                fn()
        except (asyncio.CancelledError, asyncio.TimeoutError) as ex:
            logging.warning(ex)
        except RuntimeError as ex:
            print(ex)
            raise RuntimeError(
                f"Error calling Callback function: {ex}"
            ) from ex
        except Exception as ex: # pylint: disable=W0703
            print(ex)
            logging.exception(ex, stack_info=False)


class ProviderMessaging(ProviderBase):
    """ProviderMessaging.

    Base class for All Messaging Service providers (like SMS).

    """
    provider_type = ProviderType.SMS


class ProviderIM(ProviderBase):
    """ProviderIM.

    Base class for All Message to Instant messenger providers

    """
    provider_type = ProviderType.IM
    _response = None


class ProviderPush(ProviderBase):
    """ProviderPush.

    Base class for All Message to Push Notifications.

    """
    provider_type = ProviderType.PUSH
    _response = None



"""
xmpp.

XMPP use slixmpp to send jabber messages (XMPP protocol)
"""
import logging
import ssl
from typing import (
    Union,
    Any
)
# XMPP library
from slixmpp import ClientXMPP
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.xmlstream.xmlstream import NotConnectedError
from notify.models import Actor
from notify.providers import ProviderIMBase, IM
from notify.exceptions import ProviderError

from .settings import JABBER_JID, JABBER_PASSWORD


class Client(ClientXMPP):
    """
    Client.

    Extending Class to registering session and plugins
    """
    provider = 'xmpp'
    provider_type = IM

    def __init__(self, jid, password, plugins: list = None):
        ClientXMPP.__init__(self, jid, password)
        # The session_start event will be triggered when
        # the bot establishes its connection with the server
        # and the XML streams are ready for use. We want to
        # listen for this event so that we we can initialize
        # our roster.
        self.add_event_handler("session_start", self.session_start)
        # The message event is triggered whenever a message
        # stanza is received. Be aware that that includes
        # MUC messages and error messages.
        self.add_event_handler("message", self.message)
        self.add_event_handler("disconnected", self.on_disconnect)
        self.add_event_handler("connection_failed", self.on_connection_failure)
        if isinstance(plugins, list):
            for p in plugins:
                self.register_plugin(p)
        self['xep_0030'].add_feature('echo_demo')
        self.ssl_version = ssl.PROTOCOL_SSLv3

    async def session_start(self, event):
        self.send_presence()
        try:
            await self.get_roster()
        except IqError as err:
            logging.error('There was an error getting the roster')
            logging.error(err.iq['error']['condition'])
            self.disconnect()
        except IqTimeout:
            logging.error('Server is taking too long to respond')
            self.disconnect()

    def message(self, msg):
        """
        called by slixmpp on incoming XMPP messages
        """
        if msg['type'] in ('chat', 'normal'):
            msg.reply("Thanks for sending\n%(body)s" % msg).send()

    def on_disconnect(self, event):
        if self.do_reconnections:
            self.connect()

    def on_connection_failure(self, event):
        self.log("XMPP connection failed. Try to reconnect in 5min.")
        self.schedule(
            "Reconnect after connection failure",
            60*5,
            self.on_disconnect,
            event
        )

    async def close(self):
        logging.info("Terminating XMPP session")
        self.do_reconnections = False
        self.disconnect()
        logging.info("XMPP session terminated.")


class Xmpp(ProviderIMBase):
    """
    xmpp.

    XMPP message provider
    :param username: JID Jabber
    :param password: Jabber password
    """

    provider = 'xmpp'
    provider_type = IM
    client = None
    _session = None
    _id = None
    _plugins = [
        'xep_0030',  # Service Discovery
        'xep_0199',  # XMPP Ping
        'xep_0060',  # PubSub
        'xep_0004',  # Data Forms
    ]

    def __init__(self, username: str = None, password:str = None, **kwargs):
        """
        XMPP Provider.
        """
        super(Xmpp, self).__init__(**kwargs)

        # connection related settings
        self.username = username
        if username is None:
            self.username = JABBER_JID

        self.password = password
        if password is None:
            self.password = JABBER_PASSWORD

        if self.username is None or self.password is None:
            raise RuntimeWarning(
                f'to send emails via {self.name} you need to configure username & password. \n'
                'Either send them as function argument via key \n'
                '`username` & `password` or set up env variable \n'
                'as `GMAIL_USERNAME` & `GMAIL_PASSWORD`.'
            )
        self.actor = self.username
        if 'plugins' in kwargs and isinstance(kwargs['plugins'], list):
            self._plugins = self._plugins + kwargs['plugins']

    async def test(self, jid: str = None):
        if not jid:
            jid = self.client.boundjid.bare
        try:
            rtt = await self.client['xep_0199'].ping(jid, timeout=10)
            self._logger.info(f"Success! RTT: {rtt!s}")
        except IqError as e:
            error = e.iq['error']['condition']
            self._logger.info(f'Error pinging {jid}: {error}')
        except IqTimeout:
            self._logger.info(f'No response from {jid}')
        finally:
            self.client.disconnect()

    def connect(self):
        try:
            self.client = Client(
                self.username,
                self.password,
                plugins=self._plugins
            )
            # Connect to the XMPP server and start processing XMPP stanzas
            self.client.connect()
            self.client.process(forever=False)
        except Exception as e:
            raise ProviderError(e) from e
        return self.client

    async def _send_(self, to: Actor, message: Union[str, Any], subject: str = None, **kwargs) -> Any:
        try:
            context = kwargs['context']
        except KeyError:
            context = None
        try:
            self.client.send_message(
                mto=to,
                mbody=f"{subject}: {message}",
                mtype='chat'
            )
        except NotConnectedError:
            self.log("Message NOT SENT, not connected.")
        except IqError as e:
            raise ProviderError(e) from e
        except IqTimeout as e:
            raise ProviderError(e) from e
        except Exception as e:
            raise ProviderError(e) from e

    async def close(self):
        await self.client.close()



"""
onesignal.

Using OneSignal infraestructure to send push notifications to browsers.
"""
from typing import (
    Union,
    Any
)
from collections.abc import Callable
from requests.exceptions import HTTPError
# Slack API
from slack_bolt.authorization import AuthorizeResult
from slack_bolt.async_app import AsyncApp
from slack_sdk.web.async_client import AsyncWebClient
from slack_sdk.errors import SlackApiError
# notify
from navconfig.logging import logging, loglevel
from notify.providers.abstract import ProviderIM, ProviderType
from notify.models import Actor, Channel
from notify.exceptions import ProviderError, MessageError
from .settings import (
    SLACK_APP_ID,
    SLACK_CLIENT_ID,
    SLACK_CLIENT_SECRET,
    SLACK_SIGNING_SECRET,
    # Bot information:
    SLACK_TEAM_ID,
    SLACK_BOT_TOKEN,
    SLACK_DEFAULT_CHANNEL
)


async def authorize(enterprise_id, team_id, user_id, client: AsyncWebClient, logger):
    logger.info(f"{enterprise_id},{team_id},{user_id}")
    # You can implement your own logic here
    return AuthorizeResult.from_auth_test_response(
        auth_test_response=await client.auth_test(token=SLACK_BOT_TOKEN),
        bot_token=SLACK_BOT_TOKEN,
    )


class Slack(ProviderIM):
    """
    Slack.

    param:: player_ids: a list of "players", recipients of push messages
    param:: app_id: passing an APP_ID
    """
    provider = 'slack'
    provider_type = ProviderType.IM
    blocking: bool = False

    def __init__(self, *args, **kwargs):
        """
        :param player_id: user token given by OneSignal API
        :param app_id:
        :param api_key:

        """
        self.client: Callable = None
        self.app: Callable = None
        self.conversations: list = []
        super(Slack, self).__init__(*args, **kwargs)


    async def connect(self):
        try:
            # self.app = AsyncApp(signing_secret=SLACK_SIGNING_SECRET, authorize=authorize)
            logger = logging.getLogger(__name__)
            logger.setLevel(logging.INFO)
            self.client = AsyncWebClient(token=SLACK_BOT_TOKEN, logger=logger, team_id=SLACK_TEAM_ID)
            self.conversations = await self.client.conversations_list(limit=10, team_id=SLACK_TEAM_ID)
        except Exception as err:

            self._logger.error(err)
            raise ProviderError(
                f"Error connecting to Slack API {err}"
            ) from err

    async def close(self):
        print('CLOSE')
        self.client = None
        self.app = None

    async def _send_(self, to: Actor, message: Union[str, Any], **kwargs) -> Any:
        """_send_.
        Send push notification through OneSignal API.
        """
        if isinstance(to, Channel):
            # send directly to a channel:
            channel = to.channel_id
        else:
            print('AQUI ', to, to.account)
            try:
                # getting User ID for Slack Account:
                channel = to.account.userid if to.account.provider == 'slack' else None
            except (TypeError, AttributeError, KeyError) as e:
                print('ERR ', e)
                channel = None
            try:
                channel = kwargs['channel']
            except KeyError:
                channel = SLACK_DEFAULT_CHANNEL
        msg = await self._render_(to, message, **kwargs)
        notification_body = {
            "channel": channel,
            "text": msg
        }
        try:
            # Sends it!
            response = await self.client.chat_postMessage(
                **notification_body
            )
            # print('RESPONSE ::: ', response)
            return response
            # if response.status_code == 200:
            #     print(response.body) # JSON parsed response
            #     return response
            # else:
            #     raise MessageError(
            #         f"Slack: Error sending Notification: {response.http_response}",
            #         payload=response
            #     )
        except SlackApiError as ex: # An exception is raised if response.status_code != 2xx
            raise MessageError(
                f"Slack: Error sending Notification: {ex}",
            ) from ex
        except Exception as ex:
            raise ProviderError(
                f"Slack: Exception on Slack Client: {ex}"
            ) from ex



# -*- coding: utf-8 -*-
"""
o365.

Office 365 Email-based provider
"""
from typing import (
    Union,
    Any
)
from collections.abc import Callable
# 3rd party Office 365 support
from O365 import (
    Account,
    Connection,
    MSOffice365Protocol,
    Message
)
from navconfig.logging import logging
from notify.providers.mail import ProviderEmail
from notify.models import Actor
from .settings import (
    O365_CLIENT_ID,
    O365_CLIENT_SECRET,
    O365_TENANT_ID,
    O365_USER,
    O365_PASSWORD
)




logging.getLogger('requests_oauthlib').setLevel(logging.CRITICAL)


class Office365(ProviderEmail):
    """ O365-based Email Provider
    :param client-id: Email client id
    :param client-secret: Email client secret
    :param client-email: actor email
    """
    provider = 'office365'
    blocking: bool = True

    def __init__(
            self,
            client_id: str = None,
            client_secret: str = None,
            tenant_id: str = None,
            username: str = None,
            password: str = None,
            *args,
            **kwargs
        ):
        self.account: Callable = None
        self.protocol: Callable = None
        super(Office365, self).__init__(*args, **kwargs)

        # connection related settings
        self.client_id = client_id if client_id is not None else O365_CLIENT_ID

        self.client_secret = client_secret
        if self.client_secret is None:
            self.client_secret = O365_CLIENT_SECRET

        # tenant id
        self.tenant_id = tenant_id if tenant_id is not None else O365_TENANT_ID

        # username and password:
        self.username = username
        if not self.username:
            self.username = O365_USER

        self.password = password
        if not self.password:
            self.password = O365_PASSWORD

        if self.client_id is None or self.client_secret is None:
            raise RuntimeWarning(
                f'to send emails via {self.name} you need to configure client id & client secret. \n'
                'Either send them as function argument via key id and secret or setup variables \n'
                'as `O365_CLIENT_ID` & `O365_CLIENT_SECRET`.'
            )

    def connect(self):
        """
        """
        self.protocol = MSOffice365Protocol()
        # scopes_graph = self.protocol.get_scopes_for('Mail.ReadWrite')
        scopes = ["https://graph.microsoft.com/.default"]
        if self.username is not None:
            self.account = Connection(
                credentials=(self.client_id, self.client_secret),
                auth_flow_type='credentials',
                tenant_id=self.tenant_id
            )
        else:
            self.account = Account(
                credentials=(self.client_id, self.client_secret),
                auth_flow_type='credentials',
                tenant_id=self.tenant_id,
                protocol=self.protocol
            )
            result = self.account.authenticate(scope=scopes)
            print('OFFICE635 Auth: ', result)
        try:
            return self.account
        except Exception as e:
            return e

    async def close(self):
        pass

    def _render_(self, to: Actor, subject: str = None, content: str = None, **kwargs):
        """
        """
        msg = content
        if self._template:
            templateargs = {
                "recipient": to,
                "username": to,
                "message": content,
                "content": content,
                **kwargs
            }
            msg = self._template.render(**templateargs)
        else:
            try:
                msg = kwargs['body']
            except KeyError:
                msg = content
        # email message
        if self.username is not None:
            # using basic auth instead API
            message = Message(
                auth=(self.username, self.password),
                protocol=self.protocol,
                con=self.account
            )
        else:
            message = self.account.new_message()
        message.to.add(to.account.address)
        message.subject = subject
        message.body = msg
        return message

    async def _send_(self, to: Actor, message: str, subject: str,  **kwargs):
        """
        _send_.
        Logic associated with the construction of notifications
        """
        # making email connnection
        try:
            message = self._render_(to, subject, message, **kwargs)
            status = message.send()
            logging.debug(status)
            return status
        except Exception as e:
            print(e)
            raise RuntimeError(f"{e}") from e



from navconfig import config

# Office 365
O365_CLIENT_ID = config.get('O365_CLIENT_ID')
O365_CLIENT_SECRET = config.get('O365_CLIENT_SECRET')
O365_TENANT_ID = config.get('O365_TENANT_ID')
O365_USER = config.get('O365_USER')
O365_PASSWORD = config.get('O365_PASSWORD')



from navconfig import config


# email:
EMAIL_SMTP_USERNAME = config.get('stmp_host_user')
EMAIL_SMTP_PASSWORD = config.get('stmp_host_password')
EMAIL_SMTP_PORT = config.get('smtp_port', fallback=587)
EMAIL_SMTP_HOST = config.get('stmp_host')



import os
import uuid
from pathlib import Path
from datetime import datetime
from dataclasses import InitVar
from typing import (
    Any,
    List,
    Union
)
from email.parser import Parser
from email.policy import default as policy_default
from datamodel import BaseModel, Column, Field



CONTENT_TYPES = [
    "text/plain",
    "text/html",
    "multipart/alternative",
    "application/json"
]


def auto_uuid(*args, **kwargs):
    return uuid.uuid4()

def now():
    return datetime.now()

class Account(BaseModel):
    """
    Attributes for using a Provider by an User (Actor)
    """
    provider: str = Column(required=True, default='dummy')
    enabled: bool = Column(required=True, default=True)
    address: Union[str, list] = Column(required=False, default='')
    phone: Union[str, list] = Column(required=False, default='')
    userid: str = Column(required=False, default='')

    def set_address(self, address: str):
        self.address = address


class Actor(BaseModel):
    """
    Basic Actor (meta-definition), can be an Sender or a Recipient
    """
    userid: uuid.UUID = Field(required=False, primary_key=True, default=auto_uuid)
    name: str
    account: Union[Account, List[Account]]

    def __str__(self) -> str:
        return f'<{self.name}: {self.userid}>'

Recipient = Actor
Sender = Actor

class Chat(BaseModel):
    """
    Basic configuration for chat (message-based) notifications
    """
    chat_name: str = Column(required=False)
    chat_id: str = Column(required=True, primary_key=True)

class Channel(BaseModel):
    """
    Basic configuration for Channel (group) notifications.
    """
    channel_name: str = Column(required=False)
    channel_id: str = Column(required=True, primary_key=True)

class Message(BaseModel):
    """
    Message.
    Base-class for Message blocks for Notify
    TODO:
     * template needs a factory function to find a jinja processor
     *
    """
    name: str = Column(required=True, default=auto_uuid)
    body: Union[str, dict] = Column(default=None)
    content: str = Column(required=False, default='')
    sent: datetime = Column(required=False, default=now)
    template: Path

class Attachment(BaseModel):
    """Attachement.

    an Attachment is any document attached to a message.
    """
    name: str = Column(required=True)
    content: Any = None
    content_type: str
    type: str


class BlockMessage(Message):
    """
    BlockMessage.
    Class for Message Notifications
    TODO:
     * template needs a factory function to find a jinja processor
     *
    """
    sender: Union[Actor, List[Actor]] = Column(required=False)
    recipient: Union[Actor, List[Actor]] = Column(required=False)
    content_type: CONTENT_TYPES = Column(default_factory=CONTENT_TYPES)
    attachments: List[Attachment] = Column(default_factory=list)
    flags: List[str]



class MailAttachment(Attachment):
    subject: str = Column(required=False)
    filename: str
    content_disposition: str
    attachment: Any
    size: int


class MailMessage(BlockMessage):
    """
    MailMessage.
        Dataclass for representing an Email Object.
    """

    # TODO: add validation for path-like objects using pathlib
    directory: str = Column(required=True)
    content: str = Column(required=False)
    attachments: List[MailAttachment]
    raw: InitVar = ''

    def __post_init__(self, raw: str) -> None: # pylint: disable=W0221
        msg = Parser(policy=policy_default).parsestr(raw)
        if msg:
            self.subject = msg['subject']
            self.sender = msg['from']
            self.recipient = msg['to']
            self.body = msg.get_body()
            self.attachments = []
            for part in msg.walk():
                if part.get_content_maintype() == 'text' and 'attachment' not in part.get('Content-Disposition', ''):
                    self.content = part.get_payload(decode=True).decode(part.get_param('charset', 'ascii')).strip()
                if part.get_filename() is not None:
                    attach = MailAttachment(**{
                        "name": os.path.basename(part.get_filename()),
                        "filename": part.get_filename(),
                        "content_type": part.get_content_type(),
                        "type": part.get_content_maintype(),
                        "content_disposition": part.get('Content-Disposition'),
                        "attachment": part.get_payload(decode=True),
                        'size': len(part.as_bytes())
                    })
                    self.attachments.append(attach)
        super(MailMessage, self).__post_init__()

    def getSubject(self):
        return self.subject

    def get_attachments(self):
        return self.attachments

    def get_attachments_names(self):
        return [at.filename for at in self.attachments]



from pathlib import Path
from typing import (
    Optional
)
from jinja2 import (
    Environment,
    FileSystemLoader
)


jinja_config = {
    'enable_async': False,
    'extensions': [
        'jinja2.ext.i18n'
    ]
}


class TemplateParser(object):
    """
    TemplateParser.

    This is a wrapper for the Jinja2 template engine.
    """
    def __init__(self, directory: Path, **kwargs):
        self.template = None
        self.path = directory.resolve()
        if not self.path.exists():
            raise RuntimeError(
                f'Notify: template directory {directory} does not exist'
            )
        if 'config' in kwargs:
            self.config = {**jinja_config, **kwargs['config']}
        else:
            self.config = jinja_config
        # creating loader:
        templateLoader = FileSystemLoader(
            searchpath=[str(self.path)]
        )
        # initialize the environment
        try:
            # TODO: check the bug ,encoding='ANSI'
            self.env = Environment(loader=templateLoader, **self.config)
            #compiled_path = BytesIO()
            compiled_path = str(self.path.joinpath('.compiled'))
            self.env.compile_templates(target=compiled_path, zip='deflated')
        except Exception as err:
            raise RuntimeError(
                f'Notify: Error loading Template Environment: {err}'
            ) from err

    def get_template(self, filename: str):
        self.template = self.env.get_template(str(filename))
        return self.template

    @property
    def environment(self):
        return self.env

    def render(self, filename: str, params: Optional[dict] = None) -> str:
        if not params:
            params = {}
        result = None
        try:
            self.template = self.env.get_template(str(filename))
            result = self.template.render(**params)
            return result
        except Exception as err:
            raise RuntimeError(
                f'Notify: Error rendering template: {filename}, error: {err}'
            ) from err



import logging
import importlib
from .exceptions import (
    ProviderError,
    notifyException
)

PROVIDERS = {}

class Notify(object):
    """Notify

        Factory object for getting a new Notification Provider.
    Args:
        provider (str): Name of the provider.

    Raises:
        ProviderError: when a driver cannot be loaded.
        NotSupported: when a method is not supported.
    Returns:
        ProviderBase: a Notify Provider.
    """
    def __new__(cls, provider: str, *args, **kwargs):
        _provider = None
        try:
            if not provider in PROVIDERS:
                obj = LoadProvider(provider)
                PROVIDERS[provider] = obj
            obj = PROVIDERS[provider]
            _provider = obj(*args, **kwargs)
            logging.debug(
                f':: Load Provider: {provider}'
            )
            return _provider
        except Exception as ex:
            logging.exception(
                f"Cannot Load provider {provider}: {ex}"
            )
            raise ProviderError(
                message=f"Cannot Load provider {provider}: {ex}"
            ) from ex

    @classmethod
    def provider(cls, provider, *args, **kwargs):
        try:
            if not provider in PROVIDERS:
                obj = LoadProvider(provider)
                PROVIDERS[provider] = obj
            obj = PROVIDERS[provider]
            _provider = obj(*args, **kwargs)
            logging.debug(
                f':: Load Provider: {provider}'
            )
            return _provider
        except Exception as ex:
            logging.exception(
                f"Cannot Load provider {provider}: {ex}"
            )
            raise ProviderError(
                message=f"Cannot Load provider {provider}: {ex}"
            ) from ex


def LoadProvider(provider):
    """
    loadProvider.

    Dynamically load a defined provider
    """
    try:
        # try to using importlib
        classpath = f'notify.providers.{provider}'
        module = importlib.import_module(classpath, package='providers')
        obj = getattr(module, provider.capitalize())
        return obj
    except ImportError:
        try:
            obj = __import__(classpath, fromlist=[provider])
            return obj
        except ImportError as e:
            raise notifyException(
                f'Error: No Provider {provider} was Found'
            ) from e



"""Notify Meta information."""

__title__ = 'async-notify'
__description__ = ('Library for send notifications. '
                   'simple and powerful asyncio-based library for sending notifications.')
__version__ = '0.8.11'
__author__ = 'Jesus Lara'
__author_email__ = 'jesuslara@phenobarbital.info'
__license__ = 'APACHE'



from navconfig import config

# Twitter Tweets:
TWITTER_ACCESS_TOKEN = config.get('TWITTER_ACCESS_TOKEN')
TWITTER_TOKEN_SECRET = config.get('TWITTER_TOKEN_SECRET')
TWITTER_CONSUMER_KEY = config.get('TWITTER_CONSUMER_KEY')
TWITTER_CONSUMER_SECRET = config.get('TWITTER_CONSUMER_SECRET')



import logging
from io import BytesIO
from pathlib import PurePath
from typing import (
    Union,
    Any
)
import emoji
from PIL import Image
# telegram
from aiogram import Bot, types
# from aiogram.dispatcher import Dispatcher
# from aiogram.dispatcher.webhook import SendMessage
from aiogram.utils.exceptions import (
    TelegramAPIError,
    # BotBlocked,
    BadRequest,
    MessageError,
    Unauthorized,
    NetworkError,
    ChatNotFound
)
import aiofiles
# from aiogram.utils.emoji import emojize
# from aiogram.utils.markdown import bold, code, italic, text
# TODO: web-hooks
# from aiogram.utils.executor import start_webhook
from notify.models import Actor, Chat
from notify.exceptions import notifyException
from notify.providers.abstract import ProviderIM, ProviderType
from .settings import TELEGRAM_BOT_TOKEN, TELEGRAM_CHAT_ID

aiogram_logger = logging.getLogger('aiogram')
aiogram_logger.setLevel(logging.WARNING)

class Telegram(ProviderIM):
    provider: str = 'telegram'
    provider_type = ProviderType.IM
    blocking: bool = False
    parseMode: str = 'html'  # can be MARKDOWN_V2 or HTML

    def __init__(self, *args, **kwargs):
        try:
            self.parseMode = kwargs['parse_mode']
            del kwargs['parse_mode']
        except KeyError:
            self.parseMode = 'html'
        self._bot = None
        self._info = None
        self._bot_token = None
        self._chat_id: str = None
        self._connected: bool = False
        super(Telegram, self).__init__(*args, **kwargs)
        # connection related settings
        try:
            self._bot_token = kwargs['bot_token']
        except KeyError:
            self._bot_token = TELEGRAM_BOT_TOKEN
        try:
            self._chat_id = kwargs['chat_id']
        except KeyError:
            self._chat_id = TELEGRAM_CHAT_ID


    async def close(self):
        self._bot = None
        self._connected = False

    async def connect(self, *args, **kwargs):
        # creation of bot
        try:
            self._bot = Bot(token=self._bot_token)
            self._info = await self._bot.get_me()
            self._logger.debug(
                f"🤖 Hello, I'm {self._info.first_name}.\nHave a nice Day!"
            )
            self._connected = True
        except Exception as err:
            raise notifyException(
                f"Notify: Error creating Telegram Bot {err}"
            ) from err

    def set_chat(self, chat):
        self._chat_id = chat

    def get_chat(self, **kwargs):
        # define destination
        try:
            chat = kwargs['chat_id']
            if isinstance(chat, Chat):
                self._chat_id = chat.chat_id
            del kwargs['chat_id']
        except KeyError:
            self._chat_id = TELEGRAM_CHAT_ID
        return self._chat_id

    async def _send_(
            self,
            to: Union[str, Actor, Chat],
            message: Any = None,
            subject: str = None,
            **kwargs
        ): # pylint: disable=W0221
        """
        _send_.

        Logic associated with the construction of notifications
        """
        # start sending a message:
        try:
            msg = await self._render_(to, message, subject=subject, **kwargs)
            self._logger.info(f'Messsage> {msg}')
        except Exception as err:
            raise RuntimeError(
                f'Notify Telegram: Error Parsing Message: {err}'
            ) from err
        # Parsing Mode:
        if self.parseMode == 'html':
            mode = types.ParseMode.HTML
        else:
            mode = types.ParseMode.MARKDOWN_V2
        if 'chat_id' in kwargs:
            chat_id = self.get_chat(**kwargs)
        else:
            if isinstance(to, Chat):
                chat_id = to.chat_id
            elif isinstance(to, str):
                chat_id = to
            else:
                chat_id = self._chat_id
        try:
            args = {
                'chat_id': chat_id,
                'text': msg,
                'parse_mode': mode,
                **kwargs
            }
            print(args)
            response = await self._bot.send_message(
                **args
            )
            # TODO: make the processing of response
            return response
        except Unauthorized as err:
            # remove update.message.chat_id from conversation list
            print(err)
        except ChatNotFound as err:
            # the chat_id of a group has changed, use e.new_chat_id instead
            print(err)
        except BadRequest as err:
            # handle malformed requests - read more below!
            print(err)
        except NetworkError as err:
            # handle slow connection problems
            print(err)
        except TelegramAPIError as err:
            # handle all other telegram related errors
            print(err)
        except Exception as err:
            print('ERROR: ', err)


    async def prepare_photo(self, photo):
        # Migrate to aiofile
        if isinstance(photo, PurePath):
            # is a path, I need to open that image
            img = Image.open(r"{}".format(photo))
            bio = BytesIO()
            bio.name = photo.name
            img.save(bio, img.format)
            bio.seek(0)
            return bio
        elif isinstance(photo, str):
            # its an URL
            return photo
        elif isinstance(photo, BytesIO):
            # its a binary version of photo
            photo.seek(0)
            return photo
        else:
            return None

    async def send_photo(self, photo, **kwargs):
        image = await self.prepare_photo(photo)
        if image:
            chat_id = self.get_chat()
            try:
                response = await self._bot.send_photo(
                    chat_id, photo=image, **kwargs
                )
                # print(response) # TODO: make the processing of response
                return response
            except Unauthorized as err:
                # remove update.message.chat_id from conversation list
                print(err)
            except ChatNotFound as err:
                # the chat_id of a group has changed, use e.new_chat_id instead
                print(err)
            except BadRequest as err:
                # handle malformed requests - read more below!
                print(err)
            except NetworkError as err:
                # handle slow connection problems
                print(err)
            except TelegramAPIError as err:
                # handle all other telegram related errors
                print(err)
            except Exception as err:
                print('ERROR: ', err)

    async def get_document(self, doc: Union[str, PurePath, Any]) -> Any:
        if isinstance(doc, PurePath): # Path to a File:
            if doc.exists():
                async with aiofiles.open(doc, 'rb') as f:
                    content = await f.read()
                return content
            else:
                raise FileNotFoundError(
                    f"Telegram Bot: file {doc} doesn't exists."
                )
        else:
            # TODO: check if URL or get file_id
            return doc

    async def send_document(self, document, **kwargs):
        chat_id = self.get_chat()
        doc = await self.get_document(document)
        try:
            response = await self._bot.send_document(
                chat_id,
                document=doc,
                **kwargs
            )
            # print(response) # TODO: make the processing of response
            return response
        except Unauthorized as err:
            # remove update.message.chat_id from conversation list
            print(err)
        except MessageError as err:
            print(err)
        except ChatNotFound as err:
            # the chat_id of a group has changed, use e.new_chat_id instead
            print(err)
        except BadRequest as err:
            # handle malformed requests - read more below!
            print(err)
        except NetworkError as err:
            # handle slow connection problems
            print(err)
        except TelegramAPIError as err:
            # handle all other telegram related errors
            print(err)
        except Exception as err:
            print('ERROR: ', err)

    async def get_sticker(self, sticker_id):
        if isinstance(sticker_id, dict): # getting from an sticker_set
            name = sticker_id['set']
            em = sticker_id['emoji']
            print(emoji.emojize(em))
            if isinstance(em, str) and em.startswith(':'):
                em = emoji.emojize(em)
            try:
                sticker_set = await self._bot.get_sticker_set(name)
                self._logger.debug(
                    f"Set Found: {sticker_set!r}"
                )
                st = [x for x in sticker_set['stickers'] if x['emoji'] == em]
                sticker = st[0].file_id
                return sticker
            except Exception as err:
                self._logger.warning(f'Sticker finder error: {err}')
                return None
        else:
            return 'CAACAgEAAxkBAAIuOWMze9CZzg6cQaEulHqjrcRCvBh2AAK_AgACJjFuAVdTX0Nu_LoxKgQ'

    async def send_sticker(self, sticker: Union[str, Any], **kwargs):
        chat_id = self.get_chat()
        sticker = await self.get_sticker(sticker)
        try:
            response = await self._bot.send_sticker(
                chat_id,
                sticker=sticker,
                **kwargs
            )
            # print(response) # TODO: make the processing of response
            return response
        except Unauthorized as err:
            # remove update.message.chat_id from conversation list
            print(err)
        except MessageError as err:
            print(err)
        except ChatNotFound as err:
            # the chat_id of a group has changed, use e.new_chat_id instead
            print(err)
        except BadRequest as err:
            # handle malformed requests - read more below!
            print(err)
        except NetworkError as err:
            # handle slow connection problems
            print(err)
        except TelegramAPIError as err:
            # handle all other telegram related errors
            print(err)
        except Exception as err:
            print('ERROR: ', err)

    async def get_media(self, media: Union[str, PurePath, Any]) -> Any:
        if isinstance(media, PurePath): # Path to a File:
            if media.exists():
                async with aiofiles.open(media, 'rb') as f:
                    content = await f.read()
                return content
            else:
                raise FileNotFoundError(
                    f"Telegram Bot: file {media} doesn't exists."
                )
        else:
            # TODO: check if URL or get file_id
            return media

    async def send_video(self, video: Union[str, Any], **kwargs):
        chat_id = self.get_chat()
        media = await self.get_media(video)
        try:
            response = await self._bot.send_video(
                chat_id,
                video=media,
                **kwargs
            )
            # print(response) # TODO: make the processing of response
            return response
        except Unauthorized as err:
            # remove update.message.chat_id from conversation list
            print(err)
        except MessageError as err:
            print(err)
        except ChatNotFound as err:
            # the chat_id of a group has changed, use e.new_chat_id instead
            print(err)
        except BadRequest as err:
            # handle malformed requests - read more below!
            print(err)
        except NetworkError as err:
            # handle slow connection problems
            print(err)
        except TelegramAPIError as err:
            # handle all other telegram related errors
            print(err)
        except Exception as err:
            print('ERROR: ', err)

    async def send_audio(self, audio: Union[str, PurePath, Any], **kwargs):
        chat_id = self.get_chat()
        sound = await self.get_media(audio)
        try:
            response = await self._bot.send_audio(
                chat_id,
                audio=sound,
                **kwargs
            )
            # print(response) # TODO: make the processing of response
            return response
        except Unauthorized as err:
            # remove update.message.chat_id from conversation list
            print(err)
        except MessageError as err:
            print(err)
        except ChatNotFound as err:
            # the chat_id of a group has changed, use e.new_chat_id instead
            print(err)
        except BadRequest as err:
            # handle malformed requests - read more below!
            print(err)
        except NetworkError as err:
            # handle slow connection problems
            print(err)
        except TelegramAPIError as err:
            # handle all other telegram related errors
            print(err)
        except Exception as err:
            print('ERROR: ', err)



# -*- coding: utf-8 -*-
from notify.providers.mail import ProviderEmail
from .settings import (
    EMAIL_SMTP_USERNAME,
    EMAIL_SMTP_PASSWORD,
    EMAIL_SMTP_HOST,
    EMAIL_SMTP_PORT
)

class Email(ProviderEmail):
    """
    email.

    Basic SMTP Provider.
    """
    provider = 'email'
    blocking: bool = False


    def __init__(self, hostname: str = None, port: str = None, username: str = None, password: str = None, **kwargs):
        """

        """
        self._attachments: list = []
        self.force_tls: bool = True
        self.username = None
        self.password = None

        super(Email, self).__init__(**kwargs)
        # port
        self.host = hostname
        if not self.host: # already configured
            self.host = EMAIL_SMTP_HOST
        # port
        self.port = port
        if not self.port:
            self.port = EMAIL_SMTP_PORT

        # connection related settings
        self.username = username
        if self.username is None:
            self.username = EMAIL_SMTP_USERNAME

        self.password = password
        if self.password is None:
            self.password = EMAIL_SMTP_PASSWORD

        if self.username is None or self.password is None:
            raise RuntimeWarning(
                f'to send messages via **{self._name}** you need to configure user & password. \n'
                'Either send them as function argument via key \n'
                '`username` & `password` or set up env variable \n'
                'as `EMAIL_USERNAME` & `EMAIL_PASSWORD`.'
            )
        try:
            # sent from another account
            if 'account' in kwargs:
                self.actor = kwargs['account']
            else:
                self.actor = self.username
        except KeyError:
            self.actor = self.username



"""
onesignal.

Using OneSignal infraestructure to send push notifications to browsers.
"""
from typing import (
    Union,
    Any
)
from requests.exceptions import HTTPError
from onesignal_sdk.client import AsyncClient
from onesignal_sdk.error import OneSignalHTTPError
from notify.providers.abstract import ProviderIMBase, ProviderType
from notify.models import Actor
from notify.exceptions import ProviderError
from .settings import (
    ONESIGNAL_PLAYER_ID,
    ONESIGNAL_OS_APP_ID,
    ONESIGNAL_OS_API_KEY
)



class Onesignal(ProviderIMBase):
    """
    onesignal.

    param:: player_ids: a list of "players", recipients of push messages
    param:: app_id: passing an APP_ID
    """
    provider = 'onesignal'
    provider_type = ProviderType.PUSH
    blocking: bool = False

    def __init__(self, player_ids: str = None, app_id: str = None, api_key: str = None, *args, **kwargs):
        """
        :param player_id: user token given by OneSignal API
        :param app_id:
        :param api_key:

        """
        super(Onesignal, self).__init__(*args, **kwargs)
        self.players = ONESIGNAL_PLAYER_ID if player_ids is None else player_ids
        self.os_app_id = ONESIGNAL_OS_APP_ID if app_id is None else app_id
        self.os_api_key = ONESIGNAL_OS_API_KEY if api_key is None else api_key
        self.client = None


    async def connect(self):
        try:
            self.client = AsyncClient(
                app_id=self.os_app_id,
                rest_api_key=self.os_api_key
            )
        except Exception as err:
            self._logger.error(err)
            raise ProviderError(
                f"Error connecting to OneSignal API {err}"
            ) from err


    async def _send_(self, to: Actor, message: Union[str, Any], **kwargs) -> Any:
        """_send_.
        Send push notification through OneSignal API.
        """
        notification_body = {
            'contents': {'en': message},
            'included_segments': ['Active Users'],
        }
        try:
            # Sends it!
            response = await self.client.send_notification(notification_body)
            print(response.body) # JSON parsed response
            print(response.status_code) # Status code of response
            print(response.http_response) # Original http response object.
            return response
        except OneSignalHTTPError as e: # An exception is raised if response.status_code != 2xx
            print(e)
            print(e.status_code)
            print(e.http_response.json()) # You can see the details of error by parsing original response
        except HTTPError as e:
            result = e.response.json()
            print(result)



from navconfig import config


# Amazon AWS
AWS_EMAIL_USER = config.get('aws_email_user')
AWS_EMAIL_PASSWORD = config.get('aws_email_password')
AWS_EMAIL_HOST = config.get(
    'aws_email_host', fallback='email-smtp.us-east-1.amazonaws.com'
)
AWS_EMAIL_PORT = config.get('aws_email_port', fallback=587)
AWS_EMAIL_ACCOUNT = config.get('aws_email_account')



class notifyException(Exception):
    """Base class for other exceptions."""
    code: int = None
    payload: str = None

    def __init__(self, message: str, *args, code: int = None, payload: str = None, **kwargs):
        super(notifyException, self).__init__(*args, **kwargs)
        self.args = (
            message,
            code,
        )
        self.message = message
        if code:
            self.code = code
        if payload:
            self.payload = payload

    def __str__(self):
        return f"{__name__} -> {self.message}"

    def get(self):
        return self.message


class DataError(notifyException, ValueError):
    """An error caused by invalid query input."""


class NotSupported(notifyException):
    """Not Supported functionality."""


class ProviderError(notifyException):
    """Database Provider Error."""

class MessageError(notifyException):
    """Raises when an error on Message."""

class UninitializedError(ProviderError):
    """Exception when provider cant be initialized."""

class ConnectionTimeout(ProviderError):
    """Connection Timeout Error."""



"""
Google Mail (gmail).

Using gmail library to send Email Messages.
"""
from typing import (
    Union,
    Any
)
# 3rd party gmail support
import smtplib
from gmail import GMail as GMailWorker, Message
from notify.providers.mail import ProviderEmail
from notify.exceptions import ProviderError
from notify.models import Actor
from .settings import GMAIL_USERNAME, GMAIL_PASSWORD

class Gmail(ProviderEmail):
    """
    Gmail.

        Gmail-based Email Provider.
    Args:
        :param username: Email client username
        :param password: Email client password
    """
    provider = 'gmail'
    blocking: bool = True

    def __init__(self, username: str = None, password: str = None, **kwargs):
        super(Gmail, self).__init__(**kwargs)

        # connection related settings
        self.username = username
        if username is None:
            self.username = GMAIL_USERNAME

        self.password = password
        if password is None:
            self.password = GMAIL_PASSWORD

        if self.username is None or self.password is None:
            raise RuntimeWarning(
                f'to send emails via **{self.provider}** you need to configure username & password. \n'
                'Either send them as function argument via key \n'
                '`username` & `password` or set up env variable \n'
                'as `GMAIL_USERNAME` & `GMAIL_PASSWORD`.'
            )
        self.actor = self.username

    def close(self):
        if self._server:
            try:
                self._server.close()
            except Exception as err:
                self._logger.warning(err)

    def connect(self):
        """
        connect.

        Making a connection to Gmail Servers
        """
        try:
            self._server = GMailWorker(
                self.username, self.password
            )
        except smtplib.SMTPAuthenticationError as err:
            raise Exception(
                f'Authentication Error: {err}'
            ) from err
        except Exception as err:
            raise RuntimeError(err) from err

    def _render_(self, to: Actor, content: str = None, subject: str = None, **kwargs):
        """
        """
        msg = content
        if self._template:
            self._templateargs = {
                "recipient": to,
                "username": to,
                "message": content,
                "content": content,
                **kwargs
            }
            msg = self._template.render(**self._templateargs)
        else:
            try:
                msg = kwargs['body']
            except KeyError:
                msg = content
        # email
        email = {
            'subject': subject,
            'text': msg,
            'sender': self.actor,
            'to': to.account.address,
            'html': msg
        }
        return Message(**email)

    async def _send_(self, to: Actor, message: Union[str, Any], subject: str = None, **kwargs) -> Any:
        """
        _send_.

        Logic associated with the construction of notifications
        """
        data = self._render_(to, message, subject, **kwargs)
        # making email connnection
        try:
            return self._server.send(data)
        except Exception as e:
            raise ProviderError(
                f"Gmail: Error sending Email to {to}: {e}"
            ) from e



# -*- coding: utf-8 -*-
"""
Amazon AWS Email.

Sending Emails using Amazon SMTP services
"""
from notify.providers.mail import ProviderEmail
from .settings import (
    AWS_EMAIL_USER,
    AWS_EMAIL_ACCOUNT,
    AWS_EMAIL_PASSWORD,
    AWS_EMAIL_HOST,
    AWS_EMAIL_PORT
)

class Aws(ProviderEmail):
    """ AWS-based Email Provider
    Args:
        :param username: Email client username
        :param password: Email client password
    """
    provider = 'aws_email'
    blocking: bool = False


    def __init__(self, hostname: str = None, port: str = None, username: str = None, password: str = None, *args, **kwargs):

        super(Aws, self).__init__(*args, **kwargs)

        self.username = username
        self.password = password
        self.host = hostname
        self.port = port

        try:
            self.host = kwargs['host']
        except KeyError:
            pass
        if not self.host:
            self.host = self.host = AWS_EMAIL_HOST

        if not self.port:
            self.port = AWS_EMAIL_PORT

        # connection related settings
        if self.username is None:
            self.username =     AWS_EMAIL_USER

        if self.password is None:
            self.password = AWS_EMAIL_PASSWORD

        try:
            # sent from another account
            if 'account' in kwargs:
                self.actor = kwargs['account']
            else:
                self.actor = AWS_EMAIL_ACCOUNT
        except KeyError:
            self.actor = AWS_EMAIL_ACCOUNT



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
- navigator_notify.settings
~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
from pathlib import Path
from navconfig import config, BASE_DIR

NOTIFY_DIR = Path(__file__).resolve().parent.parent

# config settings
NAVCONFIG = config
DEBUG = config.getboolean('DEBUG', fallback=True)

# Facebook Messenger
FB_USER = config.get('FB_USER')
FB_PASSWORD = config.get('FB_PASSWORD')


# TEMPLATE SYSTEM
template_dir = config.get('TEMPLATE_DIR')
if not template_dir:
    TEMPLATE_DIR = BASE_DIR.joinpath('templates')
else:
    TEMPLATE_DIR = Path(template_dir).resolve()



from navconfig import config

# OneSignail
ONESIGNAL_PLAYER_ID = config.get('ONESIGNAL_PLAYER_ID', fallback='')
ONESIGNAL_OS_APP_ID = config.get('ONESIGNAL_OS_APP_ID', fallback='')
ONESIGNAL_OS_API_KEY = config.get('ONESIGNAL_OS_API_KEY', fallback='')



# -*- coding: utf-8 -*-
"""Async-Notify.

Asyncio-based Notifications connectors for NAV.
"""
import asyncio
import pkgutil
from pathlib import Path
import uvloop
from notify.settings import TEMPLATE_DIR
from notify.templates import TemplateParser
from notify.providers.abstract import ProviderType
from .notify import PROVIDERS, Notify, LoadProvider

from .version import (
    __title__, __description__, __version__, __author__, __author_email__
)

## more information
__copyright__ = 'Copyright (c) 2020-2022 Jesus Lara'
__license__ = 'BSD'

# install uvloop and set as default loop for asyncio.
asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

__all__ = ('PROVIDERS', 'Notify', 'LoadProvider', 'ProviderType', )
TemplateEnv = None


if __name__ == "notify":
    # loading template parser:
    TemplateEnv = TemplateParser(
        directory=TEMPLATE_DIR
    )



#!/usr/bin/env python
"""Notify.

    Asynchronous library for sending notifications, used by Navigator.
See:
  https://github.com/phenobarbital/async-notify
"""
import ast
from os import path

from setuptools import find_packages, setup


def get_path(filename):
    return path.join(path.dirname(path.abspath(__file__)), filename)


def readme():
    with open(get_path('README.md'), 'r', encoding='utf-8') as rd:
        return rd.read()


version = get_path('notify/version.py')
with open(version, 'r', encoding='utf-8') as meta:
    # exec(meta.read())
    t = compile(meta.read(), version, 'exec', ast.PyCF_ONLY_AST)
    for node in (n for n in t.body if isinstance(n, ast.Assign)):
        if len(node.targets) == 1:
            name = node.targets[0]
            if isinstance(name, ast.Name) and \
                    name.id in {
                        '__version__',
                        '__title__',
                        '__description__',
                        '__author__',
                        '__license__', '__author_email__'}:
                v = node.value
                if name.id == '__version__':
                    __version__ = v.s
                if name.id == '__title__':
                    __title__ = v.s
                if name.id == '__description__':
                    __description__ = v.s
                if name.id == '__license__':
                    __license__ = v.s
                if name.id == '__author__':
                    __author__ = v.s
                if name.id == '__author_email__':
                    __author_email__ = v.s

setup(
    name=__title__,
    version=__version__,
    url='https://github.com/phenobarbital/async-notify',
    description=__description__,
    long_description=readme(),
    long_description_content_type='text/markdown',
    python_requires=">=3.8.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Environment :: Web Environment",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: AsyncIO",
    ],
    keywords=['aiogram', 'asyncio', 'aioimaplib', 'aiobotocore'],
    author='Jesus Lara',
    author_email='jesuslara@phenobarbital.info',
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'examples']),
    setup_requires=[
        "wheel==0.38.4",
        "Cython==0.29.32",
        "asyncio==3.4.3"
    ],
    install_requires=[
        'asyncio==3.4.3',
        'uvloop==0.17.0',
        'APScheduler==3.10.0',
        'aiosmtplib==2.0.0',
        'emoji==2.1.0',
        'aiogram==2.23.1',
        'pillow==9.3.0',
        'gmail==0.6.3',
        'google-auth>=2.6.0',
        'google-auth-httplib2>=0.1.0',
        'onesignal-sdk==2.0.0',
        'pyo365==0.1.3',
        'msal==1.20.0',
        'PySocks==1.7.1',
        'pyshorteners==1.0.1',
        'twilio==7.15.3',
        'tweepy==4.12.1',
        'slixmpp==1.8.3',
        "botocore==1.27.59",
        "aiobotocore==2.4.0",
        "o365==2.0.26",
        "slack_bolt==1.15.3",
        "asyncdb[default]>=2.1.30",
        "navconfig[default]>=1.0.6"
    ],
    tests_require=[
            'pytest>=5.4.0',
            'coverage',
            'pytest-asyncio',
            'pytest-xdist',
            'pytest-assume'
    ],
    test_suite='tests',
    project_urls={  # Optional
        'Source': 'https://github.com/phenobarbital/async-notify',
        'Funding': 'https://paypal.me/phenobarbital',
        'Say Thanks!': 'https://saythanks.io/to/phenobarbital',
    },
)



## for abstract email provider:
import os
import ssl
import asyncio
from abc import ABC
from typing import (
    Union,
    Any
)
from collections.abc import Callable
from email import encoders
# from email.message import EmailMessage
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.utils import formatdate
import aiosmtplib
from notify.models import Actor
from notify.exceptions import ProviderError
# abstract class
from .abstract import ProviderBase, ProviderType


class ProviderEmail(ProviderBase, ABC):
    """
    ProviderEmail.

    Base class for All Email-based providers
    """

    provider_type = ProviderType.EMAIL

    def __init__(self, *args, **kwargs):
        self.host: str = None
        self.port: int = None
        self._server: Callable = None
        super(ProviderEmail, self).__init__(*args, **kwargs)

    @property
    def user(self):
        return self.username

    async def close(self):
        if self._server:
            try:
                await self._server.quit()
            except aiosmtplib.errors.SMTPServerDisconnected:
                pass
            except Exception as err: # pylint: disable=W0703
                self._logger.exception(err, stack_info=True)
            finally:
                self._server = None

    async def connect(self, *args, **kwargs):
        """
        Make a connection to the SMTP Server
        """
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.options |= ssl.OP_NO_SSLv2
        context.options |= ssl.OP_NO_SSLv3
        context.options |= ssl.OP_NO_TLSv1
        context.options |= ssl.OP_NO_TLSv1_1
        context.options |= ssl.OP_NO_COMPRESSION
        try:
            self._server = aiosmtplib.SMTP(
                hostname=self.host,
                port=self.port,
                username=self.username,
                password=self.password,
                start_tls=True,
                tls_context=context,
                # loop=self._loop
            )
            try:
                await self._server.connect()
                self._logger.debug(f':: {self.__name__}: Connected to: {self._server}')
                try:
                    if self._server.is_ehlo_or_helo_needed:
                        await self._server.ehlo()
                except aiosmtplib.errors.SMTPHeloError:
                    pass
                await asyncio.sleep(0)
                # # making authentication:
                # await self._server.login(
                #     username=self.username,
                #     password=self.password
                # )
            except aiosmtplib.errors.SMTPAuthenticationError as err:
                raise RuntimeError(
                    f'{self.__name__} Error: Invalid credentials: {err}'
                ) from err
            except aiosmtplib.errors.SMTPServerDisconnected as err:
                raise RuntimeError(
                    f'{self.__name__} Server Disconnected: {err}'
                ) from err
        except aiosmtplib.SMTPRecipientsRefused as err:
            raise RuntimeError(
                f'{self.__name__} Error: got SMTPRecipientsRefused: {err.recipients}'
            ) from err
        except (OSError, aiosmtplib.errors.SMTPException) as e:
            raise RuntimeError(
                f'{self.__name__} Error: got {e.__class__}, {e}'
            ) from e

    def is_connected(self):
        if self._server:
            return self._server.is_connected
        else:
            return False

    def _prepare_message(self, to_address: Actor, message: Union[str, Any], content: Any): # pylint: disable=W0613
        """prepare_message.
        """
        if isinstance(content, dict):
            html = content['html']
            text = content['text']
        else:
            text = content
            html = None
        if html:
            message.add_header('Content-Type', 'text/html')
            #message.add_header('Content-Type: multipart/mixed')
            #message.add_header('Content-Transfer-Encoding: base64')
            message.attach(MIMEText(html, 'html'))
            #message.set_payload(html)
        else:
            message.add_header('Content-Type', 'text/plain')
            message.attach(MIMEText(text, 'plain'))
        return message

    async def _render_(
            self,
            to: Actor = None,
            subject: str = None,
            content: str = None,
            **kwargs
        ):
        """
        _render.

        Returns the parseable version of Email.
        """
        #TODO: add attachments
        message = MIMEMultipart('alternative')
        message['From'] = self.actor
        if isinstance(to, list):
            # TODO: iterate over actors
            message['To'] = ", ".join(to)
        else:
            message['To'] = to.account.address
        message['Subject'] = subject
        message['Date'] = formatdate(localtime=True)
        message['sender'] = self.actor
        message.preamble = subject
        if content:
            message.attach(MIMEText(content, 'plain'))
        if self._template:
            self._templateargs = {
                "recipient": to,
                "username": to,
                "message": content,
                "content": content,
                **kwargs
            }
            msg = self._template.render(**self._templateargs)
        else:
            msg = content
        message.add_header('Content-Type', 'text/html')
        #message.add_header('Content-Type', 'multipart/mixed')
        #message.add_header('Content-Transfer-Encoding', 'base64')
        message.attach(MIMEText(msg, 'html'))
        #message.set_payload(msg)
        return message

    def add_attachment(self, message, filename, mimetype='octect-stream'):
        content = None
        with open(filename, 'rb') as fp:
            content = fp.read()
        if mimetype in ('image/png'):
            part = MIMEImage(content)
        else:
            part = MIMEBase('application', 'octect-stream')
            part.set_payload(content)
            encoders.encode_base64(part)
        file = os.path.basename(filename)
        part.add_header(
            'Content-Disposition', 'attachment', filename=str(file)
        )
        message.attach(part)

    async def _send_(self, to: Actor, message: str, subject: str, **kwargs): # pylint: disable=W0221
        """
        _send.

        Logic associated with the construction of notifications
        """
        msg = await self._render_(to, subject, message, **kwargs)
        if 'attachments' in kwargs:
            for attach in kwargs['attachments']:
                self.add_attachment(
                    message=msg,
                    filename=attach
                )
        try:
            try:
                response = await self._server.send_message(msg)
                if self._debug is True:
                    self._logger.debug(response)
            except aiosmtplib.errors.SMTPServerDisconnected as err:
                raise RuntimeError(
                    f'{self.__name__} Server Disconnected {err}'
                ) from err
            return response
        except Exception as e:
            print('AQUI ERROR: ', e)
            self._logger.exception(e)
            raise ProviderError(
                f'{self.__name__} Error: got {e.__class__}, {e}'
            ) from e

    async def send(
            self,
            recipient: list[Actor] = None,
            message: Union[str, Any] = None,
            subject: str = None,
            **kwargs
        ):
        result = None
        # making the connection to the service:
        if not self._server:
            try:
                if asyncio.iscoroutinefunction(self.connect):
                    await self.connect()
                else:
                    self.connect()
            except Exception as err:
                raise ProviderError(
                    f"Error connecting to Mail Backend: {err}"
                ) from err
        # after connection, proceed exactly like other connectors.
        try:
            result = await super(ProviderEmail, self).send(recipient, message, subject=subject, **kwargs)
        except Exception as err:
            raise ProviderError(
                f"Error sending Email: {err}"
            ) from err
        return result




from typing import (
    Union,
    Any
)
from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client
from notify.providers.abstract import ProviderMessageBase, ProviderType
from notify.models import Actor
from notify.exceptions import ProviderError
from .settings import TWILIO_AUTH_TOKEN, TWILIO_ACCOUNT_SID, TWILIO_PHONE

class Twilio(ProviderMessageBase):
    provider = 'sms'
    provider_type = ProviderType.SMS
    level = ''

    def __init__(self, sid: str = None, token: str = None, **kwargs):
        """
        :param token: twilio auth token given by Twilio
        :param sid: twilio auth id given by Twilio

        """
        self._msg = None
        self.client = None
        super(Twilio, self).__init__(**kwargs)
        self.token = TWILIO_AUTH_TOKEN if token is None else token
        self.sid = TWILIO_ACCOUNT_SID if sid is None else sid

    def close(self):
        self.client = None

    def connect(self):
        """
        Verifies that a token and sid were given

        """
        if self.token is None or self.sid is None:
            raise RuntimeError(
                f'to send SMS via {self.__name__} you need to configure TWILIO_ACCOUNT_SID & TWILIO_AUTH_TOKEN in \n'
                'environment variables or send account_sid & auth_token in instance.'
            )
        self.client = Client(self.sid, self.token)

    async def _send_(self, to: Actor, message: Union[str, Any], **kwargs) -> Any:
        """
        _send.

        Send a text message using twilio
        :param to: recipient number
        :param message: message to send
        """
        try:
            data = self._render(to, message, **kwargs)
            phone = to.account['phone']
            msg = self.client.messages.create(to=phone, from_=TWILIO_PHONE, body=data)
            #print(msg)
            #print(msg.sid)
            # TODO: processing the output, adding callbacks
            return msg
        except TwilioRestException as ex:
            print(ex)
            raise ProviderError(
                f'Error Sending SMS on Twilio, current error: {ex}'
            ) from ex
